// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SB16.H) Copyright 1998 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com


#if BX_USE_SB16_SMF
#  define BX_SB16_SMF   static
#  define BX_SB16_THIS  bx_sb16.
#  define BX_SB16_THISU bx_sb16
#else
#  define BX_SB16_SMF
#  define BX_SB16_THIS  this->
#  define BX_SB16_THISU this
#endif

// some shortcuts to save typing
#define LOGFILE         BX_SB16_THIS logfile
#define BOTHLOG(x)      (x)
#define MIDILOG(x)      ((bx_options.sb16.midimode>0?x:0x7f))
#define WAVELOG(x)      ((bx_options.sb16.wavemode>0?x:0x7f))
#define MIDIDATA        BX_SB16_THIS midifile
#define WAVEDATA        BX_SB16_THIS wavefile
#define MPU             BX_SB16_THIS mpu401.
#define MPUDATAIN       MPU datain.
#define MPUDATAOUT      MPU dataout.
#define MPUCMD          MPU command.
#define MPUMIDICMD      MPU midicmd.
#define DSP             BX_SB16_THIS dsp.
#define DSPDATAIN       DSP datain.
#define DSPDATAOUT      DSP dataout.
#define MIXER           BX_SB16_THIS mixer.
#define EMUL            BX_SB16_THIS emuldata.
#define EMULIN          EMUL datain.
#define EMULOUT         EMUL dataout.

// If the buffer commands are to be inlined:
#define BUFINL inline

// the hardware properties. These can be changed via a DSP command
#define BX_SB16_IO      0x220       // IO base address of DSP & FM part
#define BX_SB16_IOLEN   16          // number of addresses covered
#define BX_SB16_IOMPU   0x330       // IO base address of MPU402 part
#define BX_SB16_IOMPULEN 4          // number of addresses covered
// IRQ, DMA8bit and DMA16bit for the DSP part. Note, these
// are changeable at runtime in mixer registers 0x80 and 0x81
#define BX_SB16_IRQ     BX_SB16_THIS currentirq
//#define BX_SB16_IRQMPU  9           // IRQ for the MPU401 part - not supported!
#define BX_SB16_DMAL    BX_SB16_THIS currentdma8
#define BX_SB16_DMAH    BX_SB16_THIS currentdma16

#define BX_SB16_MIX_REG  0x90       // total number of mixer registers

typedef Bit32u (*bx_sb16_read_t)();
typedef void   (*bx_sb16_write_t)(Bit32u value);

     // The array containing an instrument/bank remapping
struct bx_sb16_ins_map {
  Bit8u oldbankmsb, oldbanklsb, oldprogch;
  Bit8u newbankmsb, newbanklsb, newprogch;
};

// This is the class for the input and
// output FIFO buffers of the SB16

class bx_sb16_buffer {
public:

  BUFINL bx_sb16_buffer(void);
  BUFINL ~bx_sb16_buffer(void);
  BUFINL void init(int bufferlen);
  BUFINL void reset();

      /* These functions return 1 on success and 0 on error */
  BUFINL Boolean put(Bit8u data);    // write one byte in the buffer
  BUFINL Boolean puts(char *data, ...);  // write a formatted string to the buffer
  BUFINL Boolean get(Bit8u *data);   // read the next available byte
  BUFINL Boolean getw(Bit16u *data); // get word, in order lo/hi
  BUFINL Boolean getw1(Bit16u *data);// get word, in order hi/lo
  BUFINL Boolean full(void);         // is the buffer full?
  BUFINL Boolean empty(void);        // is it empty?

  BUFINL void flush(void);           // empty the buffer
  BUFINL int bytes(void);            // return number of bytes in the buffer
  BUFINL Bit8u peek(int ahead);      // peek ahead number of bytes
 
      /* These are for caching the command number */
  BUFINL void newcommand(Bit8u newcmd, int bytes);   // start a new command with length bytes
  BUFINL Bit8u currentcommand(void); // return the current command
  BUFINL void clearcommand(void);    // clear the command
  BUFINL Boolean commanddone(void);  // return if all bytes have arrived
  BUFINL Boolean hascommand(void);   // return if there is a pending command
  BUFINL int commandbytes(void);     // return the length of the command


private:
  Bit8u *buffer;
  int head,tail,length;
  Bit8u command;
  Boolean havecommand;
  int bytesneeded;
  };

// The actual emulator class. It intercepts accesses to the
// SB16 IO ports and writes the raw midi stream to /dev/midi00
class bx_sb16_c {
public:

  bx_sb16_c(void);
  ~bx_sb16_c(void);
  BX_SB16_SMF void init(bx_devices_c *);

      /* DMA input and outbut, 8 and 16 bit, have to be public */
  BX_SB16_SMF void   dma_write8(Bit8u *data_byte);
  BX_SB16_SMF void   dma_read8(Bit8u *data_byte);
  BX_SB16_SMF void   dma_write16(Bit16u *data_word);
  BX_SB16_SMF void   dma_read16(Bit16u *data_word);

      // IRQ, DMA8 and DMA16 also public, they're read in devices.cc
  int currentirq;
  int currentdma8;
  int currentdma16;

private:

  FILE *logfile;
  FILE *midifile,*wavefile;     // the output files or devices
  struct {
    bx_sb16_buffer datain, dataout, command, midicmd;
    Boolean uartmode, irqpending;
    int banklsb[16],bankmsb[16],program[16];   // patch translation lists
  } mpu401;

  struct {
    bx_sb16_buffer datain, dataout;
    Bit8u resetport;                           // last value written to the reset port
    Bit8u timeconstant,speaker,prostereo;      // properties of the sound input/output
    Bit16u blocklength, samplerate;            //
    Boolean irqpending;                        // Is an IRQ pending (not ack'd)
    Boolean midiuartmode;                      // Is the DSP in MIDI UART mode
    // Properties of the current DMA transfer:
    // dmamode= 0: no transfer, 1: single-cycle transfer, 2: auto-init DMA
    // dmabits= 8 or 16
    // dmafifo= ??  Bit used in DMA command, no idea what it means...
    // dmaoutput= 0: input, 1: output
    // dmabps= bytes per sample; =(dmabits/8)*(dmastereo+1)
    // dmastereo= 0: mono, 1: stereo
    // dmasigned= 0: unsigned data, 1: signed data
    // dmahighspeed= 0: normal mode, 1: highspeed mode (only SBPro)
    int dmamode, dmabits, dmafifo, dmaoutput, dmabps;
    int dmastereo, dmasigned, dmahighspeed;
    Bit16u dmacount;    // bytes remaining in this transfer
  } dsp;

  struct {
    Bit8u regindex;
    Bit8u reg[BX_SB16_MIX_REG];
  } mixer;

  struct {
    bx_sb16_buffer datain, dataout;
    bx_sb16_ins_map remaplist[256];
    int remaps;
  } emuldata;

  int timer_handle;
  int current_timer;

  BX_SB16_SMF void midi_timer (void *);
  BX_SB16_SMF Bit32u gettimer (void);


  bx_devices_c *devices;

  BX_SB16_SMF void   writelog(int loglevel, const char *str, ...);
  BX_SB16_SMF void   writemidi(const Bit8u value);

      /* the MPU 401 part of the emulator */
  BX_SB16_SMF Bit32u mpu_status();                   // read status port   3x1
  BX_SB16_SMF void   mpu_command(Bit32u value);      // write command port 3x1
  BX_SB16_SMF Bit32u mpu_dataread();                 // read data port     3x0
  BX_SB16_SMF void   mpu_datawrite(Bit32u value);    // write data port    3x0

      /* The DSP part */
  BX_SB16_SMF void   dsp_reset(Bit32u value);        // write to reset port 2x6
  BX_SB16_SMF Bit32u dsp_dataread();                 // read from data port 2xa
  BX_SB16_SMF void   dsp_datawrite(Bit32u value);    // write to data port  2xa
  BX_SB16_SMF Bit32u dsp_bufferstatus();             // read buffer status  2xc
  BX_SB16_SMF Bit32u dsp_status();                   // read dsp status     2xe
  BX_SB16_SMF Bit32u dsp_irq16ack();                 // ack 16 bit IRQ      2xf
  BX_SB16_SMF void   dsp_dma(Bit8u command, Bit8u mode, Bit16u length, Bit8u comp);
                                                     // start a DMA transfer
  BX_SB16_SMF void   dsp_sample(Bit8u value);        // receive one byte of a sample

      /* The mixer part */
  BX_SB16_SMF Bit32u mixer_readdata(void);
  BX_SB16_SMF void   mixer_writedata(Bit32u value);
  BX_SB16_SMF void   mixer_writeregister(Bit32u value);
  BX_SB16_SMF void   set_irq_dma();

      /* The emulator ports to change emulator properties */
  BX_SB16_SMF Bit32u emul_read (void);               // read emulator port
  BX_SB16_SMF void   emul_write(Bit32u value);       // write emulator port

      /* several high level sound handlers */
  BX_SB16_SMF void   writemidicommand(Boolean force);// flush the midi buffer
  BX_SB16_SMF void   midiremapprogram(int channel);  // remap program change
  BX_SB16_SMF void   writedeltatime(Bit32u deltatime, FILE *midifile);
                                                // write in delta time coding

  BX_SB16_SMF void   initmidifile(FILE *midifile);   // Write midi file header
  BX_SB16_SMF void   finishmidifile(FILE *midifile); // write track length etc.
  BX_SB16_SMF void   initwavefile(FILE *wavefile);   // Write wave file header
  BX_SB16_SMF void   finishwavefile(FILE *wavefile); // write track length etc.

      /* The port IO multiplexer functions */

  BX_SB16_SMF Bit32u read_handler(void *this_ptr, Bit32u address, unsigned io_len);
  BX_SB16_SMF void   write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len);

#if !BX_USE_SB16_SMF
  Bit32u read(Bit32u address, unsigned io_len);
  void   write(Bit32u address, Bit32u value, unsigned io_len);
#endif
  };

#if BX_USE_SB16_SMF
extern bx_sb16_c bx_sb16;
#endif
