program mpuremap;
{ MPU 401 bank/program change remap load
  for the SB16 patch to Bochs.
  Compile with Borland Pascal 5.0 or newer }

{
  This program is Copyright (C) by Josef Drexler.
  It may be redistributed according to the GNU public
  license, stated in the file COPYING.
  This program is provided as is, without any warranty.
}

{ $DEFINE DEBUG}

const
  emulport = $333;	{ Port 0x333 is the emulator port }

procedure usage;
begin
  writeln;
  writeln('Usage: MPUREMAP #  to display info string number #');
  writeln('    or MPUREMAP filename  to load a patch remap file "filename"');
  writeln;
  writeln('The patch remap file contains lines in the format');
  writeln('OldBankMSB OldBankLSB OldProgram NewBankMSB NewBankLSB NewProgram');
  writeln('Any value of 255 means ignore it. Comment lines may start with "#".');
  writeln;
end;

procedure testemul(value, index : byte);
var
  i : byte;
begin
  i := port[emulport];
  {$IFNDEF DEBUG}
  if i <> value then
    begin
      writeln('Bochs emulator/SB16 patch not found (', index, ')!');
      halt;
    end;
  {$ENDIF}
end;

procedure showemul(i : byte);
var
  j : byte;
begin
  port[emulport] := 5;   { 5 means dump info }
  port[emulport] := i;   { Argument to command 5; info number }
  testemul(254, 2);
  write('Emulator info string ', i, ': ');
  repeat
    j := port[emulport];
    if j = 255 then
      break;
    write(char(j));
  until j = 10;
end;

var
  linenr,
  i, j : integer;
  p : byte;
  trans : array[1..6] of integer;
  str : string;

begin
  writeln('MPU 401 bank/program change remap loader for the');
  writeln('SB16 patch to Bochs.');

  { Check command line}
  if paramcount = 0 then
    usage;

  { Check emulator status }
  port[emulport] := 0;
  testemul(254, 1);

  if paramcount = 0 then
    halt;
  str := paramstr(1);

  if str[1] in ['0'..'9'] then
    begin
      { Show emulator info }
      val(str, i, j);
      showemul(i);
      halt;
    end;
  { Load remap list from file }
  linenr := 0;
  assign(input, paramstr(1));
  reset(input);
  port[emulport] := 7;   { 7 reset remap list }
  testemul(254, 3);
  while not eof do
    begin
      inc(linenr);
      readln(str);
      if length(str) < 1 then
	continue;
      if str[1] = '#' then
	continue;
      for i := 1 to length(str) do
	case upcase(str[i]) of
	  'X' : if (i > 1) and (str[i - 1] = '0') then
		  begin
		    str[i - 1] := ' ';
		    str[i] := '$';
		  end;
	  ',',
	  #9 : str[i] := ' ';
	end;
      str := str + ' ';
      for i := 1 to 6 do
	begin
	  p := pos(' ', str);
	  val(copy(str, 1, p - 1), trans[i], j);
	  if j <> 0 then
	    begin
	      writeln('Parse error in value ', i, ' on line ', linenr,
		      ' of file ', paramstr(1), '. Line ignored.');
	      break;
	    end;
	  while (str[p] = ' ') and (p < length(str)) do
	    inc(p);
	  str := copy(str, p, 255);
	end;
      if j <> 0 then
	continue;
      { Load the remap into the emulator, command 4 }
      port[emulport] := 4;   { 4 load remap }
      testemul(254, 4);

      for i := 1 to 6 do
	port[emulport] := trans[i];

      testemul(6, 5);	     { test receipt of 6 arguments }
    end;
  showemul(3);
  close(input);
end.
